# -*- coding: utf-8 -*-

"""
britney.middleware
~~~~~~~~~~~~~~~~~~

:copyright: (c) 2013 by Arnaud Grausem
:license: BSD see LICENSE for details
"""

from functools import partial


class Middleware(object):
    """
    """

    def add_headers(self, environ, *headers):
        """
        """
        for header in headers:
            environ['spore.headers'].append(header)

    def __call__(self, environ):
        """
        """
        if hasattr(self, 'process_request'):
            environ.setdefault('spore.headers', [])
            self.process_request(environ)
        if hasattr(self, 'process_response'):
            callback = partial(self.process_response)
            return callback



# -*- coding: utf-8 -*-

"""
britney.middleware.auth
~~~~~~~~~~~~~~~~~~~~~~~

"""

from base64 import b64encode
from functools import partial

from . import Middleware


def _basic_auth(username, password):
    """
    """
    return 'Basic ' + b64encode(
        ('%s:%s' % (username, password)).encode('latin1')
    ).strip().decode('latin1')


class Auth(Middleware):
    """
    """

    def needs_auth(self, environ):
        return environ['spore.authentication']

    def __call__(self, environ):
        """
        """
        if self.needs_auth(environ):
            super(Auth, self).__call__(environ)


class Basic(Auth):
    """
    """

    def __init__(self, username, password):
        self.username = username
        self.password = password

    def process_request(self, environ):
        """
        """
        header = ('Authorization', _basic_auth(self.username, self.password))
        self.add_headers(environ, header)


class Proxy(Basic):
    """
    """
    
    def __init__(self, username, password):
        super(Proxy, self).__init__(username, password)


    def process_request(self, environ):
        """
        """
        header = ('Proxy-Authorization', _basic_auth(self.username,
            self.password))
        self.add_headers(environ, header)


class ApiKey(Auth):
    """
    """

    def __init__(self, key_name, key_value, *args):
        self.key = key_name
        if callable(key_value):
            self.value = key_value(*args)
        else:
            self.value = key_value

    def process_request(self, environ):
        """
        """
        header = (self.key, self.value)
        self.add_headers(environ, header)



# -*- coding: utf-8 -*-

"""
"""

import re
from requests import Request
from requests.compat import quote


class RequestBuilder(object):
    """
    """

    def __init__(self, env):
        self.env = env
        self.path_info = env['PATH_INFO']
        self.query_string = env.get('QUERY_STRING', '')

        # building path and query with params submitted
        for param, value in env['spore.params']:
            self.path_info, p_changed = re.subn(':%s' % param, value, 
                    self.path_info)
            self.query_string, q_changed = re.subn(':%s' % param, value,
                    self.query_string)
            if not p_changed and not q_changed:
                self.query_string += '&%s=%s' % (param, value)

        # correct query string
        if self.query_string.startswith('&'):
            self.query_string = self.query_string[1:]

    @property
    def application_uri(self):
        """
        """
        uri = self.env['wsgi.url_scheme'] + '://'

        if self.env.get('HTTP_HOST', ''):
            uri += self.env['HTTP_HOST']
        else:
            if self.env['spore.userinfo']:
                uri += self.env['spore.userinfo'] + '@'
            
            uri += self.env['SERVER_NAME']

            if self.env['wsgi.url_scheme'] == 'https':
                if self.env['SERVER_PORT'] != '443':
                    uri += ':' + self.env['SERVER_PORT']
            else:
                if self.env['SERVER_PORT'] != '80':
                    uri += ':' + self.env['SERVER_PORT']

        return uri + quote(self.env['SCRIPT_NAME'] or '/')
    
    @property
    def uri(self):
        """
        """
        uri = self.application_uri

        path_info = quote(self.path_info, safe='/=;,')
        if not self.env['SCRIPT_NAME']:
            uri += path_info[1:]
        else:
            uri += path_info

        if self.query_string:
            uri += '?' + quote(self.query_string, safe='&=;,')

        return uri

    def headers(self):
        """
        """
        return dict(self.env.get('spore.headers', ()))

    def data(self):
        """
        """
        return self.env['spore.payload'] or {}


    def __call__(self):
        """
        """
        request = Request(
                method=self.env['REQUEST_METHOD'],
                url=self.uri,
                data=self.data(),
                headers=self.headers()
        )
        return request.prepare()



# -*- coding: utf-8 -*-

"""
britney
~~~~~~~

The Python implementation of SPORE (Specifications to a POrtable Rest
Environment)

This project is based on spyre
"""

import requests
from .core import Spore


def spyre(spec_uri, base_url=None):
    """
    """
    if spec_uri.startswith('http'):
        func = _new_from_url
    else:
        func = _new_from_file

    api_description = func(spec_uri)
    api_description.update({'base_url': base_url})

    return Spore(**api_description)

def _new_from_file(spec_uri):
    """
    """
    import json

    with open(spec_uri, 'r') as spec_file:
        spec = json.loads(spec_file.read())

    return spec


def _new_from_url(spec_uri):
    """
    """
    response = requests.get(spec_uri)
    return response.json()



# -*- coding: utf-8 -*-

"""
britney.errors
~~~~~~~~~~~~~~

Core errors made by britney
"""


class SporeClientBuildError(Exception):

    def __init__(self, client_errors, method_errors, *args, **kwargs):
        super(SporeClientBuildError, self).__init__(*args, **kwargs)
        self.errors = client_errors
        if method_errors:
            self.errors['methods'] = method_errors
        
        
class SporeMethodBuildError(Exception):

    def __init__(self, errors, *args, **kwargs):
        super(SporeMethodBuildError, self).__init__(*args, **kwargs)
        self.errors = errors


class SporeMethodCallError(Exception):

    def __init__(self, cause, *args, **kwargs):
        self.expected_values = kwargs.pop('expected', [])
        super(SporeMethodCallError, self).__init__(*args, **kwargs)
        self.cause  = cause

class SporeMethodStatusError(Exception):
    """
    """
    pass



# -*- coding: utf-8 -*-

"""
britney.middleware.format
~~~~~~~~~~~~~~~~~~~~~~~~~

:copyright: (c) 2013 - Arnaud Grausem
:licence: BSD see LICENSE for details
"""

import abc
import json

from . import Middleware


class Format(Middleware):
    """
    """

    __metaclass__ = abc.ABCMeta


    @abc.abstractmethod
    def dump(self):
        """
        """
        pass

    @abc.abstractmethod
    def load(self):
        """
        """
        pass
    
    @abc.abstractproperty
    def accept_type(self):
        """
        """
        pass

    @abc.abstractproperty
    def content_type(self):
        """
        """
        pass

    def process_request(self, environ):
        headers = [self.accept_type]
        if environ['spore.payload']:
            environ['spore.payload'] = self.dump(environ['spore.payload'])
            headers.append(self.content_type)
        self.add_headers(environ, *headers)

    def process_response(self, response):
        response.data = response.json()


class Json(Format):
    """
    """

    def dump(self, data):
        return json.dumps(data)

    def load(self, data):
        return json.loads(data)

    @property
    def content_type(self):
        return ('Content-Type', 'application/json')

    @property
    def accept_type(self):
        return ('Accept', 'application/json')



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from setuptools import setup
from setuptools import find_packages

with open('README.rst') as readme:
    long_description = readme.read()

with open('requirements.txt') as requirements:
    lines = requirements.readlines()
    libraries = [lib for lib in lines if not lib.startswith('-')]
    dependency_links = [link.split()[1] for link in lines if 
            link.startswith('-f')]

setup(
    name='britney',
    version='0.1',
    author='Arnaud Grausem',
    author_email='arnaud.grausem@gmail.com',
    maintainer='Arnaud Grausem',
    maintainer_email='arnaud.grausem@gmail.com',
    url='https://github.com/agrausem/britney',
    license='PSF',
    description=u'Python implementation of SPORE',
    long_description=long_description,
    packages=find_packages('src'),
    package_dir={'': 'src'},
    download_url='http://pypi.python.org/pypi/britney',
    install_requires=libraries,
    dependency_links=dependency_links,
    keywords=['SPORE', 'REST Api', 'client'],
    entry_points={},
    classifiers = (
        'Development Status :: 3 - Alpha',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
    )
)



# -*- coding: utf-8 -*-

"""
britney.core
~~~~~~~~~~~~

Core objects that builds the whole SPORE client, based on informations defined
in a SPORE description. See https://github.com/SPORE/specifications/blob/master/spore_description.pod for
more informations about SPORE descriptions
"""


import requests
from requests.compat import is_py2
from requests.compat import urlparse

from . import errors
from .request import RequestBuilder


class Spore(object):
    """ Base class generating at run-time the Spore HTTP Client to a REST
    Environnment.

    :param name: name of the REST Web Service (required)
    :param base_url: base url of the REST Web Service (required)
    :param authority: authority that maintains the SPORE description of the
    REST Web Service (defaults to '')
    :param formats: a list of accepted formats (eg: json, xml) (defaults to
    None
    :param authentication: boolean for a global authentication check (defaults
    to None)
    :param methods: a dict containing the methods information that will
    instantiate a :py:class:`~britney.core.SporeMethod` (required)
    :param meta: meta informations about the description and the service
    """

    def __new__(cls, *args, **kwargs):
        spec_errors = {}

        if not kwargs.get('name', ''):
            spec_errors['name'] = 'A name for this client is required'

        if not kwargs.get('methods', {}):
            spec_errors['methods'] = 'One method is required to create the client'

        if not kwargs.get('base_url', ''):
            spec_errors['base_url'] = 'A base URL to the REST Web Service is '
            'required'

        setattr(cls, 'middlewares', [])

        method_errors = {}
        for method_name, method_description in kwargs['methods'].items():
            try:
                method = SporeMethod(
                    name=method_name, 
                    base_url=kwargs['base_url'],
                    middlewares=cls.middlewares,
                    global_authentication=kwargs.get('authentication', None),
                    global_formats=kwargs.get('formats', None),
                    **method_description
                )
            except errors.SporeMethodBuildError as method_error:
                method_errors[method_name] = method_error
            else:
                setattr(cls, method_name, method)
            
        if spec_errors or method_errors:
            raise errors.SporeClientBuildError(spec_errors,
                    method_errors)

        return super(Spore, cls).__new__(cls, *args, **kwargs)

    def __init__(self, name='', base_url='', authority='', formats=None, 
            version='', authentication=None, methods=None, meta=None): 
        self.name = name
        self.authority = authority
        self.base_url = base_url
        self.version = version
        self.authentication = authentication

        self.formats = [] if formats is None else formats
        self.meta = {} if meta is None else meta

    def __repr__(self):
        return '<Spore [{}]>'.format(self.name)

    def enable(self, middleware, **kwargs):
        """ Enables a middleware on the client object

        :param middleware: a middleware class
        :param kwargs: parameters to instantiate the middleware
        """
        self.enable_if(lambda request: True, middleware, **kwargs)

    def enable_if(self, predicate, middleware, **kwargs):
        """ Enables a middleware on the client object only if the predicate is
        true

        :param predicate: a function that takes the request as argument and
        returns a boolean
        :param middleware: a middleware class
        :param kwargs: parameters to instantiate the middleware
        """
        if not callable(middleware):
            raise ValueError(middleware)

        self.middlewares.append((predicate, middleware(**kwargs)))


class SporeMethod(object):
    """ A method that handles request and response on the REST Web Service.

    :param name: name of the method (required)
    :param api_base_url: the base url for the REST Web Service (required)
    :param method: the http method to use (eg: 'GET', 'POST', etc) (required)
    :param path: the path to the wanted resource (eg: '/api/users') (required)
    :param required_params: a list parameters that will be used to build the
    url path (defaults to None)
    :param optional_params: a list of parameters that wille be used to build
    the request query
    :param expected_status: a list of expected status for the reponse (defaults
    to None).
    :param description: a short description for the method (defaults to '')
    :param middlewares: a list of the middlewares that will be applied to the
    request and the reponse. See :py:class:~`britney.middleware.BaseMiddleware`
    :param authentication: boolean to set authentication on the method. This
    param replaces the global authentication parameter set for the whole client
    on this particular method (defaults to None)
    :param formats: a list of formats accepted by the REST Web Service for this
    method. This parameter replaces the global formats set whole client
    (defaults to None)
    :param base_url: a specific base url for this method. This parameter 
    replaces the base api url set for the whole client (defaults to '')
    :param documentation: documentation for this method
    :param global_authentication: a boolean that enables authentication for the
    whole client
    :param global_formats: a list of formats accepted for the whole client.
    """

    PAYLOAD_HTTP_METHODS = ('POST', 'PUT', 'PATCH')
    HTTP_METHODS = PAYLOAD_HTTP_METHODS + ('GET', 'TRACE', 'OPTIONS', 'DELETE',
            'HEAD')

    def __new__(cls, *args, **kwargs):
        method_errors = {}

        if not kwargs.get('method', ''):
            method_errors['method'] = 'A method description should define the HTTP '
            'Method to use'

        if not kwargs.get('path', ''):
            method_errors['path'] = 'A method description should define the path to '
            'the wanted resource(s)'

        if not kwargs.get('name', ''):
            method_errors['name'] = 'A method description should define a name'

        if not kwargs.get('api_base_url', '') \
                and not kwargs.get('base_url', ''):
            method_errors['base_url'] = 'A method description should define a base '
            'url if not defined for the whle client'

        if method_errors:
            raise errors.SporeMethodBuildError(method_errors)
        
        documentation = kwargs.get('documentation', '')
        description = kwargs.get('description', '')

        setattr(cls, __doc__, documentation if documentation else description)

        return super(SporeMethod, cls).__new__(cls, *args, **kwargs)

    def __init__(self, name='', api_base_url='', method='', path='', 
            required_params=None, optional_params=None, expected_status=None,
            required_payload=False, description='', authentication=None,
            formats=None, base_url='', documentation='', middlewares=None,
            global_authentication=None, global_formats=None):

        self.name = name
        self.method = method
        self.path = path
        self.description = description
        self.required_payload = required_payload

        self.base_url = base_url if base_url else api_base_url
        self.formats = formats if formats else global_formats

        self.required_params = required_params if required_params else []
        self.optional_params = optional_params if optional_params else []
        self.middlewares = middlewares
        self.expected_status = expected_status if expected_status else []

        self.headers = []

        if authentication is None:
            self.authentication = global_authentication \
                if global_authentication is not None else False
        else:
            self.authentication = authentication

    def __repr__(self):
        return '<SporeMethod [{}]>'.format(self.name)

    def base_environ(self):
        """ Builds the base environment dictionnary describing the request to
        be sent to the REST Web Service. See https://github.com/SPORE/specifications/blob/master/spore_implementation.pod
        for more informations about the keys defined here. You can also check
        the WSGI environnment keys http://wsgi.readthedocs.org/en/latest/definitions.html
        """

        parsed_base_url = urlparse(self.base_url)

        def script_name(path):
            return path.rstrip('/')

        def userinfo(parsed_url):
            if parsed_url.username is None:
                return ''
            return '{0.username}:{0.password}'.format(parsed_url)


        def server_port(parsed_url):
            if not parsed_url.port:
                if parsed_url.scheme == 'http':
                    return 80
                elif parsed_url.scheme == 'https':
                    return 443
            return parsed_url.port

        path = self.path.split('?')
        if len(path) > 1:
            path_info, query_string = path
        else:
            path_info, query_string = path[0], ''
 
        return {
            'REQUEST_METHOD': self.method,
            'SERVER_NAME': parsed_base_url.hostname,
            'SERVER_PORT': server_port(parsed_base_url), 
            'SCRIPT_NAME': script_name(parsed_base_url.path),
            'PATH_INFO': path_info,
            'QUERY_STRING': query_string,
            'HTTP_USER_AGENT': 'britney',
            'spore.expected_status': self.expected_status,
            'spore.authentication': self.authentication,
            'spore.params': '',
            'spore.payload': '',
            'spore.errors': '',
            'spore.format': self.formats,
            'spore.userinfo': userinfo(parsed_base_url),
            'wsgi.url_scheme': parsed_base_url.scheme,
        }

    def build_payload(self, data):
        """
        """

        if not data and self.required_payload:
            raise errors.SporeMethodCallError('Payload is required for '
                'this function')

        return data
    
    def build_params(self, **kwargs):
        """ Check aguments passed to call method and build the spore
        parameters value
        """

        req_params = frozenset(self.required_params)
        all_params = req_params | frozenset(self.optional_params)
        passed_args = is_py2 and kwargs.viewkeys() or kwargs.keys()

        # nothing to do here
        if not all_params and not passed_args:
            return []

        # some required parameters are missing
        if not req_params.issubset(passed_args):
            raise errors.SporeMethodCallError('Required parameters are missing', 
                    expected=req_params - passed_args)
        
        # too much arguments passed to func
        if (passed_args - all_params):
            raise errors.SporeMethodCallError('Too much parameter',
                    expected=passed_args - all_params)
        
        return list(is_py2 and kwargs.viewitems() or kwargs.items())

    def check_status(self, response):
        """ Checks response status in fact of the *expected_status*
        attribute

        :param response: the response from the REST service
        :type response: requests.Response
        :raises: ~britney.errors.SporeMethodStatusError
        """
        status = int(response.status)
        checklist = [status == expected_status for expected_status
                in self.expected_status]
        if checklist and not any(checklist):
            raise errors.SporeMethodStatusError()


    def __call__(self, **kwargs):
        """ Calls the method with required parameters
        """

        hooks = []
        data = kwargs.pop('payload', None)

        environ = self.base_environ()
        environ.update({
            'spore.payload': self.build_payload(data),
            'spore.params': self.build_params(**kwargs)
        })


        for predicate, middleware in self.middlewares:
            if predicate(environ):
                callback = middleware(environ)
                if callback is not None:
                    if isinstance(callback, requests.Response):
                        return callback
                    hooks.append(callback)

        prepared_request = RequestBuilder(environ)

        with requests.session() as session:
            response = session.send(prepared_request(), verify=True)

        # self.check_status(response)
        map(lambda hook: hook(response), hooks)

        return response
